#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "badpairs.h"
#include "admutils.h"
#include "mcio.h"  
#include "mcmcpars.h"  

#define WVERSION   "250" 
// if eglistname, aglistname not given "Case"  "Control" assumed

#define EVAL 10
#define AVAL 11

#define MAXFL  50   
#define MAXSTR  512

char *trashdir = "/var/tmp" ;

int qtmode = NO ;
int verbose = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 

double plo = .001 ;
double phi = .999 ;

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indivname = NULL ;
char *aglistname  = NULL ;
char *eglistname  = NULL ;
char *outputname = NULL ;
char *fliplistname = NULL ;  // reference allele wrong
FILE *ofile ;

void readcommands(int argc, char **argv) ;

int mkeglist(char **eglist);
void seteglist(char *eglistname, char **eglist, 
  int numeg, int val, int mode) ;


void cntit(double *xc, SNP *c1, SNP *c2) ;

double lddip(double *xc) ;
void flipallele(SNP *cupt)  ;
void setfliplist(char *flistname, char **flist, int *numfl) ;

int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  char **aglist ;
  int  lsnplist, lindlist, numeg, numag ;
  int i,j,k; 
  SNP *cupt, *cupt1, *cupt2 ;
  Indiv *indx ;
  char str[MAXSTR] ;

  int nindiv = 0, e, f, lag=1  ;
  int n, n0, n1, g, chrom, gender ;
  int *cnt0,  *cnt1 ;
  int ec0, ec1, ac0, ac1 ;
  double xc[9], xd[4], xc2[9] ;
  double ychi ;
  char *fliplist[MAXFL] ;
  int numfl ;
  int nignore, numrisks = 1;

 readcommands(argc, argv) ;

  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, NULL, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;
  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  ZALLOC(eglist, numindivs, char *) ; 
  ZALLOC(aglist, numindivs, char *) ; 
  numeg = mkeglist(eglist) ;

  seteglist(eglistname, eglist,  numeg, EVAL,0) ;
  seteglist(aglistname, aglist,  numeg, AVAL,1) ;

  setfliplist(fliplistname, fliplist, &numfl) ;
  for (i=0; i<numfl; i++)  {
   k = snpindex(snpmarkers, numsnps, fliplist[i]) ;
   if (k<0) 
    fatalx("(fliplist) %s not found\n", fliplist[i]) ;
   cupt = snpmarkers[k] ;
   flipallele(cupt) ;
  }


  ZALLOC(cnt0, numeg, int) ;
  ZALLOC(cnt1, numeg, int) ;

  printf("%20s %3s", "SNP_ID", "C") ;
  for (i=0; i<numeg; i++) {  
   strcpy(str, eglist[i]) ;
   str[9] = '\0' ;
   printf(" %9s",str) ;
  }
  printf(" %9s","PopA") ;
  printf(" %9s","PopB") ;
  printf("\n") ; 
  for (j=0; j<numsnps; j++) {  
   cupt = snpmarkers[j] ;

   ivzero(cnt0, numeg) ;
   ivzero(cnt1, numeg) ;
   ivzero(cupt -> popA_nn, 2) ;
   ivzero(cupt -> popB_nn, 2) ;

   chrom = cupt -> chrom ;

   n = cupt -> ngtypes ;
   ec0 = ec1 = 0 ;
   ac0 = ac1 = 0 ;
   for (i=0; i<n; i++) {
    indx = indivmarkers[i] ; 
    k = indxindex(eglist, numeg,  indx->egroup) ;
    if (k<0) continue ;
    g = getgtypes(cupt, i) ;
    if (g<0) continue ;
    n0 = g ;  
    n1 = 2-g ;

    if ((chrom==23) && ((indx->gender=='M'))) {  
     n0 /= 2 ;
     n1 /= 2 ;
    }
     cnt0[k] += n0 ;
     cnt1[k] += n1 ;
     if (indx -> affstatus == EVAL) {  
      ec0 += n0 ;
      ec1 += n1 ;
     }
     if (indx -> affstatus == AVAL) {  
      ac0 += n0 ;
      ac1 += n1 ;
     }
   }
   printf("%20s %3d", cupt->ID, chrom) ;
   for (k=0; k<numeg; k++) {  
    printf(" %4d %4d",cnt1[k], cnt0[k]) ;
   }
   printf (" %4d %4d %4d %4d", ac1, ac0, ec1, ec0) ;
   printf("\n") ;
   cupt -> popA_nn[1] = ac1 ;
   cupt -> popA_nn[0] = ac0 ;
   cupt -> popB_nn[1] = ec1 ;
   cupt -> popB_nn[0] = ec0 ;
  }
  printsnps(outputname, snpmarkers, numsnps, 
    indivmarkers, NO, NO) ;
  return 0 ;
}
int mkeglist(char **eglist)
{
  Indiv *indx ;
  int i, k, numeg=0 ;
  for (i=0; i<numindivs; i++) { 
   indx = indivmarkers[i] ;
    k = indxindex(eglist, numeg,  indx->egroup) ;
    if (k<0)   { 
     eglist[numeg] = strdup(indx->egroup) ;
     ++numeg ;
    }
  }
  return numeg ;
}

void checkgeno() 
{
  int i, k, x ;
  int n, nbad, ngood ;
  SNP *cupt ;

  nbad = ngood = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;
   if (n<=0) continue ;
   for (k=0; k<n; k++) {
    x = getgtypes(cupt, k) ;
    if (x>2) {  
     printf("bad genotype: %s %4d %4d\n", cupt->ID, k,x) ;
     putgtypes(cupt, k, -1) ;  
     ++nbad ;
    }
    else  {
     ++ngood ;
    }
   }
  }
  printf("checkgeno. good: %d  bad:  %d\n",ngood, nbad) ;
/**
  if (nbad>0) fatalx("checkgeno fails\n") ;
*/
}
void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "aglistname:", &aglistname) ;
   getstring(ph, "eglistname:", &eglistname) ;
   getstring(ph, "poplistname:", &eglistname) ;
   getstring(ph, "fliplistname:", &fliplistname) ;
   getstring(ph, "output:", &outputname) ;

}

void setfliplist(char *flistname, char **flist, int *numfl)
{
  FILE *fgfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, num=0 ;

  *numfl = 0; 
  if (flistname == NULL) return ;
  openit(flistname, &fgfile, "r") ;
  while (fgets(line, MAXSTR, fgfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') continue ;
   if (num>=MAXFL) 
  fatalx("too many flips\n") ;
   flist[num] = strdup(sx) ;
   ++num ;
   freeup(spt, nsplit) ;
  }
  fclose(fgfile) ;
  *numfl = num ;
}

void seteglist(char *eglistname, char **eglist, int numeg, 
  int val, int defaultmode)
// defaultmode only matters if eglistname = NULL
{
  FILE *egfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx = NULL ;
  Indiv *indx ;
  int nsplit, i ;

  if ((eglistname == NULL) || (strcmp(eglistname, "NULL") == 0))  {
   if (defaultmode==0) sx = strdup("Case") ;
   if (defaultmode==1) sx = strdup("Control") ;
   for (i=0; i<numindivs; i++) {
    indx = indivmarkers[i] ;
    if (strcmp(indx->egroup, sx) == 0)
     indx -> affstatus = val ;
   }
   free(sx) ;
   return ;
  }
  openit(eglistname, &egfile, "r") ;
  while (fgets(line, MAXSTR, egfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   for (i=0; i<numindivs; i++) {
    indx = indivmarkers[i] ;
    if (strcmp(indx->egroup, sx) == 0)
     indx -> affstatus = val ;
   }
   freeup(spt, nsplit) ;
  }
  fclose(egfile) ;
}


void flipallele(SNP *cupt) 
// flip reference, variant counts
{
   int i, n, g ;
   n = cupt -> ngtypes ;
   for (i=0; i<n; i++) {  
    g = getgtypes(cupt, i) ;
    if (g<0) continue ;
    putgtypes(cupt, i, 2-g) ;
   }
}
