
double dprime(int *a1, int*a2, int n) 
{
/** 
 a1, a2 are categories and contain small integers 
 computes Lewontin's dprime statistic for multialle locus
 Ref: Hedrick Genetics 117 (1987) pp 331-341 
*/  
 int max1, max2, len, a,b, i  ;
 double *x, t, dd ;
 double *f1, *f2 ;

 ivmaxmin(a1, n, &max1, NULL) ;
 ivmaxmin(a2, n, &max2, NULL) ;
 len = (max1+1)*(max2+1) ;
 ZALLOC(x, len, double) ;
 ZALLOC(f1, max1+1, double) ;
 ZALLOC(f2, max2+1, double) ;
 for (i=0; i<n; i++) {
  a = a1[i] ;
  b = a2[i] ;
  ++x[a*(max2+1)+b] ;
  ++f1[a] ;
  ++f2[b] ;
 }
 t = asum(f1,max1+1) ; vst(f1,f1,1.0/t,max1+1) ;
 t = asum(f2,max2+1) ; vst(f2,f2,1.0/t,max2+1) ;
 t = asum(x,len) ; vst(x,x,1.0/t,len) ;
 
 t = 0.0 ;
 for (a=0; a<=max1; ++a) {
  if (f1[a] == 0.0) continue ;
  for (b=0; b<=max2; ++b) {
   if (f2[b] == 0.0) continue ;
   dd = lewont(f1[a],f2[b],x[a*(max2+1)+b]) ;

/**
   printf("zzlewont %9.3f %9.3f %9.3f %9.3f\n", 
   f1[a],f2[b],x[a*(max2+1)+b],dd ) ;
*/

   t += f1[a]*f2[b]*fabs(dd) ;
  }
 }
 if (verbose==YES)  {
  for (i=0; i<n; i++) {
   a = a1[i] ;
   b = a2[i] ;
   printf("yy1 %4d %4d %4d\n",i,a,b) ;
  }
  for (a=0; a<=max1; ++a) {
   if (f1[a] == 0.0) continue ;
   for (b=0; b<=max2; ++b) {
    if (f2[b] == 0.0) continue ;
    dd = lewont(f1[a],f2[b],x[a*(max2+1)+b]) ;
    printf("yy2 %d %d %9.3f %9.3f %9.3f %9.3f\n",a,b,f1[a],f2[b],x[a*(max2+1)+b],dd) ;
   }
  }

 }
 free(x) ;
 free(f1) ;
 free(f2) ;
 return t ;
}
double lewont(double p, double q, double x) 
/* d(ij) statistic */
{
   double d1, dmax ;
   d1 = x - p*q ;
   if (d1<0.0) {
    dmax = MIN(p*q, (1.0-p)*(1.0-q)) ;
   }
   if (d1>0.0) {
    dmax = MIN(p*(1.0-q), q*(1.0-p)) ;
   }
   if (d1==0.0) return 0.0 ;  
   if (dmax==0.0) fatalx("(lewont) bad freq %9.3f %9.3f\n",p,q) ; 
   return d1/dmax ;
}


