#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <nicklib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "admutils.h"

extern double a1, b1 ;
extern double thxp0, thxp1, thxp2 ;

extern Indiv **indivmarkers ; 
extern SNP **snpmarkers ;

static int debug = YES ;

void calcX(SNP **snpm, int i, int j, double *X, int numstates) ;

void
mk2from3ml(double *xd, double *xc, double p, double pp) ;

double
mk2from2(double *xd, double *xc) ;

/** 
 a crude EM estimation 
 that ignores linkage on a chromosome 
*/

void
 emsimple(SNPMC **snpmcmarkers, 
 int numsnps, int numindivs, int numiter, 
 double *zstatmax, double *zstatmin) 
{
  int i, j, indiv, n, iter, g ;
  int a, t, ghalf ;
  int numstates, ignore ;
  double theta, lambda, y, ylike, ybase, xa, xb ;
  double xt, xs,  p, pp, q, qq;
  double eps = 1.0e-10 ;
  SNP *cupt, *tcupt ;
  SNPMC *cumcpt ;
  Indiv *indx ;
  double gg[3], tt[3], X[3] ;
  double *u0, *u1, *v0, *v1, *x0, *x1 ;
  double dd[4], dd2[4], ddx[4] ;

  double ncase, ncontrol ;
  double *zstat, *z1case, *z1control, *z2case, *z2control, *w1, *w2 ;
  double tmp, zmax, zmin ;
  double aa1, bb1 ;
  int nbug = 0 ;

  if (numiter == 0) {  
   *zstatmax = *zstatmin = 0.0 ;
   return ;
  }

  ZALLOC(zstat, numsnps, double) ;
  ZALLOC(z1case, numsnps, double) ;
  ZALLOC(z1control, numsnps, double) ;
  ZALLOC(z2case, numsnps, double) ;
  ZALLOC(z2control, numsnps, double) ;
  ZALLOC(w1, numsnps, double) ;
  ZALLOC(w2, numsnps, double) ;

  ZALLOC(u0, numindivs, double) ;
  ZALLOC(u1, numindivs, double) ;
  ZALLOC(v0, numindivs, double) ;
  ZALLOC(v1, numindivs, double) ;
  
 ncase = ncontrol = 0 ;
 for (indiv = 0; indiv < numindivs; ++indiv) {  
   indx = indivmarkers[indiv] ;
   if (indx -> ignore) continue ;
   if (indx -> affstatus)  ++ncase ;
   else ++ncontrol ;
 }


 printf("\n##SCORES FROM EXPECTATION_MAXIMIZATION ALGORITHM ITERATIONS\n");
 printf("##%25s %8s\n","Iteration_Num","Score");
  for (iter =1; iter <= numiter; ++iter) {  
   ylike = 0.0 ;

   vclear(u0, b1, numindivs) ;
   vclear(u1, a1, numindivs) ;
   vzero(v0,  numindivs) ;
   vzero(v1,  numindivs) ;

 for (indiv = 0; indiv < numindivs; ++indiv) {  
   indx = indivmarkers[indiv] ;

   if (iter==1) indx -> theta_mode = a1/(a1+b1) ;

   if (indx -> ignore) continue ;
   xt = p = indx->theta_mode ; 
   xs = q = 1-p ;
   ylike += (a1-1.0)*log(p) + (b1-1.0)*log(q) ;

   aa1 = thxp1 + thxp0*xt  ;
   bb1 = thxp2 + thxp0*xs  ;

   v0[indiv] = bb1 ;
   v1[indiv] = aa1 ;

   if (iter==1) indx -> Xtheta_mode = aa1/(aa1+bb1) ;

   p = indx->Xtheta_mode ; 
   q = 1-p ;

   ylike += (aa1-1.0)*log(p) + (bb1-1.0)*log(q) ;
 }


   vzero(z1case,numsnps) ;
   vzero(z2case,numsnps) ;
   vzero(z1control,numsnps) ;
   vzero(z2control,numsnps) ;

  for (j=0; j<numsnps; j++) {  
    cupt = snpmarkers[j] ;  
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;
    if (cupt -> isrfake) continue ;
    n = cupt -> ngtypes ;
    if (n==0) continue ;
    cumcpt = snpmcmarkers[j] ;
    vzero(cumcpt->xc, 9) ;
    vzero(cumcpt->xd, 4) ;

    p  = cupt -> adm_PopA_freq ;
    pp = cupt -> adm_PopB_freq ;

    if (iter == 1) {  
      p = MAX(p, 0.01) ;
      p = MIN(p, 0.99) ;

      pp = MAX(pp, 0.01) ;
      pp = MIN(pp, 0.99) ;

    }

    xa = ddx[0] = cupt -> popA_nn[0]  ;
    xb = ddx[1] = cupt -> popA_nn[1]  ;
    if (xa>=0.0) {
     ylike += (xa*log(p) + xb*log(1-p)) ;
    }
    else vzero(ddx,2) ;

    xa = ddx[2] = cupt -> popB_nn[0]  ;
    xb = ddx[3] = cupt -> popB_nn[1]  ;
    if (xa>=0.0) {
     ylike += (xa*log(pp) + xb*log(1-pp)) ;
    }
    else vzero(ddx+2,2) ;

   for (indiv=0; indiv<numindivs; ++indiv)  {
    indx = indivmarkers[indiv] ;
    gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;
    if (ignore==YES) continue  ;
    x0 = u0 ;
    x1 = u1 ;
    if ((cupt -> chrom) == 23) {
     x0 = v0 ;
     x1 = v1 ;
    }
    g = getgtypes(cupt, indiv) ;
    ++nbug ;
    if (g<0) continue ;
    sett1(tt, theta, numstates) ;
    calcX(snpmarkers, indiv, j, X, numstates) ;
    gg[2] = 0.0 ;
    vvt(gg, X, tt, numstates) ;
    y = asum(gg, numstates) ;
    vst(gg, gg, 1.0/y, numstates) ;
    ylike += log(y) ; 

/* now count ethnic a genotype g events */
    ghalf = g/2 ;
    for (a=0; a<numstates; a++) {  
     if (numstates==3) { 
      t = 3*a + g ;
      cumcpt -> xc [t]  += gg[a] ;
     }  
     else {
      t = 2*a + ghalf ;
      cumcpt -> xd [t]  += gg[a] ;
     }
    }
    x0[indiv] += (numstates-1) * gg[0] ;
    x0[indiv] += (numstates-2) * gg[1] ;
    x1[indiv] += 2 * gg[2] ;
    x1[indiv] += gg[1] ;

    tmp = 2*gg[2] + gg[1] ;
    tmp /= (double) (numstates-1)  ;

    tmp -= theta ;

   if (indx -> affstatus)  {  
    z1case[j] += tmp ;
    z2case[j] += tmp*tmp ;
   }
   else {
    z1control[j] += tmp ;
    z2control[j] += tmp*tmp ;
   }


   } /* end indiv loop  */

   mk2from3ml(dd, cumcpt -> xc, p, pp) ;
   mk2from2(dd2, cumcpt -> xd) ;
   vvp(dd, dd2,  dd, 4) ;
   vvp(dd, dd,  ddx, 4) ;
   vsp(dd, dd, 1.0e-20, 4) ;

   xa = dd[0] ; xb = dd[1] ;
   y = xa/(xa+xb) ;

   cupt -> adm_PopA_freq = y ;

   xa = dd[2] ; xb = dd[3] ;
   y = xa/(xa+xb) ;


   cupt -> adm_PopB_freq = y ;


  } /* end marker loop */
  for (indiv = 0; indiv < numindivs; ++indiv) {  
   indx = indivmarkers[indiv] ;
   if (indx -> ignore) continue ;
   xa = u0[indiv]  ;
   xb = u1[indiv]  ;
   y = xb/(xa+xb) ;
   indx -> theta_mode =  y ;
   xa = v0[indiv]  ;
   xb = v1[indiv]  ;
   y = xb/(xa+xb) ;
   indx -> Xtheta_mode =  y ;
  }
  if (iter==1) ybase = ylike ; 
  ylike -= ybase ;
  printf("emsimple iter: %6d %14.3f\n", iter, ylike) ;
 }

/** carry out a t-test */
  if ( (ncase >0 ) && (ncontrol > 0)) {  
   vst(z1case, z1case, 1.0/ncase, numsnps) ;
   vst(z2case, z2case, 1.0/ncase, numsnps) ;
   vvt(w2, z1case, z1case, numsnps) ;
   vvm(z2case, z2case, w2, numsnps) ;
   vst(z1control, z1control, 1.0/ncontrol, numsnps) ;
   vst(z2control, z2control, 1.0/ncontrol, numsnps) ;
   vvt(w2, z1control, z1control, numsnps) ;
   vvm(z2control, z2control, w2, numsnps) ;
   vvm(w1, z1case, z1control, numsnps) ; /* diff (mean)  */

   vst(z2case, z2case, 1.0/ncase, numsnps) ;
   vst(z2control, z2control, 1.0/ncontrol, numsnps) ;
   vvp(w2, z2case, z2control, numsnps) ;
   vsp(w2, w2, 1.0e-10, numsnps) ;

   vsqrt(w2, w2, numsnps) ;
   vvd(zstat, w1, w2, numsnps) ;
 }

 vmaxmin(zstat, numsnps, &zmax, &zmin) ;
 if (zstatmax != NULL) *zstatmax = zmax ;
 if (zstatmin != NULL) *zstatmin = zmin ;
 


  free(u0) ;
  free(u1) ;
  free(v0) ;
  free(v1) ;

  free(w2) ;
  free(w1) ;
  free(z2control) ;
  free(z2case) ;
  free(z1control) ;
  free(z1case) ;
  free(zstat) ;

}

void
mk2from3ml(double *xd, double *xc, double p, double pp) 
/* like mk2from3 but no random sample */
{
   double q, qq  ;
   double p1, p2 ;
   double y  ;

   int  c ;
   int  e1, e2, b1, b2, n1, n2 ;
   int  k, t ;
   double a, aa, b, bb, x1, x2, xx ;

   q = 1-p ; qq = 1-pp ;


   vzero(xd, 4) ;
   
   for (k=0; k<9; k++)  {  
     e1 = e2 = b1 = b2 = n1 = n2 = 0 ;
    switch (k)  {  
     case 0:  
       e1 = 0 ; b1 = 1; n1 = 2;  
       y = q*q ;
       break ;
     case 1:  
       e1 = 0 ; b1 = 0; n1 = 1;  
       e2 = 0 ; b2 = 1; n2 = 1;  
       y = 2*q*p ;
       break ;
     case 2:  
       e1 = 0 ; b1 = 0; n1 = 2;  
       y = p*p ;
       break ;
     case 3:  
       e1 = 0 ; b1 = 1; n1 = 1;  
       e2 = 1 ; b2 = 1; n2 = 1;  
       y = q*qq ;  
       break ;
     case 4:  
       y = p*qq + pp*q ;  
       break ;
     case 5:  
       e1 = 0 ; b1 = 0; n1 = 1;  
       e2 = 1 ; b2 = 0; n2 = 1;  
       y = p*pp ;  
       break ;
     case 6:  
       e1 = 1 ; b1 = 1; n1 = 2;  
       y = qq*qq;
       break ;
     case 7:  
       e1 = 1 ; b1 = 0; n1 = 1;  
       e2 = 1 ; b2 = 1; n2 = 1;  
       y = 2*qq*pp ;
       break ;
     case 8:  
       e1 = 1 ; b1 = 0; n1 = 2;  
       y = pp*pp ;
       break ;
    }
    t = 2*e1 + b1 ;  xd[t] += (double) n1*xc[k] ;
    t = 2*e2 + b2 ;  xd[t] += (double) n2*xc[k] ;
   }

   p1 = p*qq ; p2 = pp*q ;  
   xx =  xc[4] ;  

   x1 = (xx* p1)/(p1+p2) ;
   x2 = xx - x1 ;

   e1 = 0; b1=0; 
   e2 = 1; b2=1;

   t = 2*e1 + b1 ;  xd[t] += x1 ;
   t = 2*e2 + b2 ;  xd[t] += x1 ;

   e1 = 0; b1=1; 
   e2 = 1; b2=0;
   t = 2*e1 + b1 ;  xd[t] += x2 ;
   t = 2*e2 + b2 ;  xd[t] += x2 ;


}
