#include  <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>  
#include <time.h> 

#include <nicklib.h>
#include <getpars.h> 

int verbose = NO ;
int wgflag = YES ; 
// default invalid genotypes not written out
#define MAXIND 10000 
#define MAXSTR 512
#define MAXFF 20

char *iname = NULL ;
char *oname = NULL ;

char *indnames[MAXIND] ; 
char *pops[MAXIND] ; 
double  **gams ;  
int geno[MAXIND] ;

int readgamsx(char *nams) ;
void writegams (char *oname, int num) ;
void readcommands(int argc, char **argv) ;

double xpest(double **gg, int *gobs, int *na, int *nb,  
 int neq, double *ppa, double *ppb) ;

void xcalc(int x, double pa, double pb, double *X) ;

int   
main  (int argc , char **argv) 
{
  int n, j, g, t, doest = YES  ;
  int na[2], nb[2] ; 
  double pa, pe ;
  double ggr[3], *gg, y ;

  gams  = initarray_2Ddouble(MAXIND, 3, 0.0) ;

  readcommands(argc, argv) ;
  n = readgamsx(iname) ; 
  ivzero(na, 2) ; ivzero(nb, 2) ;

  t = intsum(geno, n) ; 
  if (t==0) doest = NO ;
  if (t==2*n) doest = NO ;
  
  if (doest) {
   xpest(gams, geno, na, nb, n, &pa, &pe) ;
   for (j=0; j<n; j++) {  
     g = geno[j] ; 
     if (g<0) continue ;
     gg = gams[j] ;
     xcalc(geno[j], pa, pe, ggr) ;
     vvt(gg, gg, ggr, 3) ;
     bal1(gg,3) ;
   }
  }
  writegams(oname, n) ;

}

void readcommands(int argc, char **argv) 

{
  int i ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int seed = 0  ;

  while ((i = getopt (argc, argv, "i:o:V")) != -1) {

    switch (i)
      {

      case 'i':
	iname = strdup(optarg) ;
	break;

      case 'o':
	oname = strdup(optarg) ;
	break;

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }
}
int readgamsx(char *gammfile) 
{

  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, j, num=0 ;
  FILE *fff ;
  double *gg ;
  int g, ind;  

  openit(gammfile, &fff, "r") ;

  while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit==0) continue ;
   sx  = spt[0] ;

   if (sx[0] == '#') { 
    freeup(spt, nsplit) ;
    continue ;
   }

   if (num>=MAXIND) fatalx("too many individuals\n") ;
   ind = num ;
   indnames[num] = strdup(sx) ;
   gg = gams[num] ;          
   gg[0] = atof(spt[1]) ;
   gg[1] = atof(spt[2]) ;
   gg[2] = atof(spt[3]) ;
   bal1(gg, 3) ;
   
   g = atoi(spt[5]) ;
   pops[num] = strdup(spt[4]) ;
   geno[num] = g ;
   freeup(spt, nsplit) ;
   ++num ;
  }
  fclose(fff) ;
  return num ;
}
void writegams (char *oname, int num) 

{
  int k ; 
  double *gg ;
  FILE *fff ;

  openit(oname, &fff, "w") ;
  for (k=0; k<num; ++k) {  
   gg = gams[k] ;
   fprintf(fff, "%20s", indnames[k]) ;
   fprintf(fff, " %9.3f", gg[0]) ;
   fprintf(fff, " %9.3f", gg[1]) ;
   fprintf(fff, " %9.3f", gg[2]) ;
   fprintf(fff, " %20s", pops[k]) ;
   fprintf(fff, " %4d", geno[k]) ;
   fprintf(fff, "\n") ;
  }

  fclose(fff) ;



}



void xcalc(int x, double pa, double pb, double *X)
{

  double PopA_freq, PopB_freq;
  static int ncall=0 ;

  ++ncall ;
  vclear(X, 1.0, 3) ;
  

  if (x<0) {  
   return ; 
  }

  PopA_freq   =  pa ;
  PopB_freq =  pb ;

  if ((PopA_freq<0) || (PopB_freq<0)) 
    fatalx("marker freqs not set\n") ;

    switch(x) {
      
    case 2:
      X[0] = PopA_freq * PopA_freq;
      X[1] = PopA_freq * PopB_freq;
      X[2] = PopB_freq * PopB_freq;
      break;
      
    case 1:
      X[0] = 2 * PopA_freq * (1 - PopA_freq) ;
      X[1] = PopB_freq * (1 - PopA_freq) + 
	(1 - PopB_freq) * PopA_freq ;
      X[2] =  2 * PopB_freq * (1 - PopB_freq) ;
      break;
      
    case 0:
      X[0] = (1 - PopA_freq) * (1 - PopA_freq);
      X[1] = (1 - PopA_freq) * (1 - PopB_freq);
      X[2] = (1 - PopB_freq) * (1 - PopB_freq);
      break;
  
    default:
      fatalx("(xcalc) bad genotype %d\n", x) ;
      break;
    }
}


