#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>  
#include <nicklib.h>  
#include "admutils.h" 

#define MAXSTR 2048
#define MAXFIELD 200

/** 
 rank snps in admix 
 in information order 
*/

struct node {  
 int snum ;  
 double xp1 ;
 double xp2 ;
 int npt ;
 double score ;
 struct node *last ;
 struct node *next ;
}  ;


#define VERSION "2001"

int verbose=NO  ;
static int debug=NO  ;

#define MAXNUM 200000
void getdata(FILE *ifile, int *pos, double *gendis,
  double *p0, double *p1, char **pplace, int *num, 
  char **tags, int *forceit) ;
void getndlist(int *nlist, int *nsize, double dis, int nsnp, int k)  ;
void getnlist(int *nlist, int *nsize, int dis, int nsnp, int k)  ;
void mknlist2(int *nbuff, int *nsize, int *xmark)  ;
double mutxx(double *dd, int m , int n) ;
void m3(double *xout, double a0, double a1, double pf) ;
void m2(double *x, double *a, double *b, double scal)  ; 

double mutinf(int *a, int n, double pf) ;
double oldmutinf(int *a, int n, double pf) ;
double mutinfd(int *a, int n, double pf) ;

double
 mutxhmmd(double *p0, double *p1, double *dis, int n, int kmark, double theta, double lambda)  ;
// code in ranksubs.c 

void cklist(char *str, struct node *pfix ) ; /* debug */
void remlist(struct node *pfix, struct node *head)  ;
void inslist(struct node *pfix, struct node *head)  ;

static int  *pos ;
static double *gendis ;
static double *p1 ;
static double *p2 ;

double theta, lambda = 0.0 ;

int usephys = NO ;
char *tagforce = NULL ; /* if tag matches force score high */

#define KMAX 20 
#define KKMAX 8

int kmax = KKMAX ;
int main(int argc, char **argv) 
{
   char c ;

   char *iname = NULL ;
   char *oname = NULL ;
   FILE *ofile, *ifile ;

   double popf[2] ; 
   double pf3[3] ;
   double tt[2] ;
   double cfreq = .21 ;
   double ybase, y, tdis ;
   int nsnp ;
   int i, k, j, kk ;
   int alist[KMAX] ;
   double *xinf ;
   double *tinf ; 
   int *xind, *xmark ;
   int *nbuff, nsize ;
   int *outlist ;
   int noclose = 100000 ;
   double noclosegd = 0.0 ;  /* cM */
   double adblim  = 4.0  ;
   struct node *head, *pt, *ptnext, *ps ;
   struct node *xlist ;
   int nout ;
   char **pplace, **tags ;
   double flat=0.0 ; 
   int *forceit ;
  

  ZALLOC(pos, MAXNUM, int) ;
  ZALLOC(gendis, MAXNUM, double) ;
  ZALLOC(p1,  MAXNUM, double) ;
  ZALLOC(p2,  MAXNUM, double) ;
  ZALLOC(pplace,  MAXNUM, char *) ;
  ZALLOC(forceit, MAXNUM, int) ;
  ZALLOC(tags,  MAXNUM, char *) ;

  ivclear(forceit, NO, MAXNUM) ;

  tagforce = strdup("") ;

  while ((c = getopt(argc, argv, "a:i:o:w:x:y:f:t:l:k:vVp")) != EOF)

    switch (c) {

    case 'a':
      cfreq = atof(optarg);
      break;

    case 'l':
      lambda = atof(optarg);
      break;

    case 'f':
      flat = atof(optarg);
      break;

    case 't':
      tagforce = strdup(optarg);
      break;

    case 'i':
      iname = strdup(optarg);
      break;

    case 'o':
      oname = strdup(optarg);
      break;

    case 'k':
      kmax = atoi(optarg);
      break;

    case 'x':
      noclose = atoi(optarg);
      break;

    case 'w':
      noclosegd = atof(optarg);
      break;

    case 'y':
      adblim  = atof(optarg);  /* cMorgans  or phys distance*/
      break;

    case 'p':
      usephys = YES ;
      break;

    case 'v':
      printf("version: %s\n", VERSION) ;
      break;

    case 'V':
      verbose = YES ;
      break;

    default:
      fatalx("bad argument flag\n") ;

    }

   theta = cfreq ; 
  
   ifile = stdin ;
   ofile = stdout ;

   if (iname != NULL) ifile = fopen(iname,"r") ;
   if (oname != NULL) ofile = fopen(oname,"w") ;

   fprintf(ofile, "## cfreq: %9.3f\n",cfreq) ;
   if (flat>0.0) {
    fprintf(ofile, "## flat: %9.3f\n",flat) ;
   }
   fprintf(ofile, "## noclose: %d  adblim:  %9.3f noclosegd: %9.3f\n",  
    noclose, adblim, noclosegd) ;
   fprintf(ofile,"##%10s %9s %9s %2s  %9s %9s %9s %15s %s\n",
    "pos", "gendis", 
     "score", "n", "rawscore", "p(C)", "p(A)", "origin", "rs tag") ;

   popf[0] = cfreq ;
   popf[1] = 1.0-cfreq ;

   m2(pf3, popf, popf, 1.0) ;
   printf("## entropy of ancestry (1 chromosome) %9.3f\n", entrop(popf,2)) ;
   ybase = entrop(pf3,3) ;
   printf("## entropy of ancestry (2 chromosomes) %9.3f\n", ybase) ;

   if (ifile==NULL) fatalx("bad ifile\n") ;
   getdata(ifile, pos, gendis, p1, p2, pplace, &nsnp, 
    tags, forceit) ;

   if (flat>0.0) {  
    vst(p1, p1, 1.0-flat, nsnp) ;
    vsp(p1, p1, flat*0.5, nsnp) ;
    vst(p2, p2, 1.0-flat, nsnp) ;
    vsp(p2, p2, flat*0.5, nsnp) ;
   }

   fclose(ifile) ;
   printf("## number of snps: %d\n",nsnp) ;
   printf("## noclose:  %d\n",noclose) ;
   printf("## adblim:  %9.3f\n",adblim) ;

   if (nsnp == 0) fatalx("no SNP data\n") ;
   if (usephys) floatit(gendis, pos, nsnp) ;

   ZALLOC(xinf, nsnp, double) ;
   ZALLOC(tinf, nsnp, double) ;
   ZALLOC(xind, nsnp, int) ;
   ZALLOC(nbuff, nsnp, int) ; /* used for neighbor list */
   ZALLOC(xlist, nsnp, struct node) ;
   ZALLOC(xmark, nsnp, int) ;  /* 1 if used */
   ZALLOC(outlist, nsnp, int) ;
   nout = 0 ; 
   for (i=0; i<nsnp; i++) {  
    alist[0] = i ;
    xinf[i] = y = mutinf(alist,1,cfreq) ;
    if (forceit[i]) xinf[i] = 2.0 ;
    if (verbose) 
     printf("## mutual inf: %5d %9.3f  %15d\n",i,y,pos[i]) ;
   }
   vst(tinf, xinf, -1.0, nsnp) ;
   sortit(tinf, xind, nsnp) ;

   for (i=0; i<nsnp; i++) {  
    pt = &xlist[i] ;
    pt -> snum = i ;
    pt -> npt = 0 ;
    pt->score = xinf[i] ;
    pt-> next = NULL ;
    pt-> last = NULL ;
    pt -> xp1 = p1[i] ;
    pt -> xp2 = p2[i] ;
   }

/** 
 build linked list forwards 
*/
   k = xind[0] ;
   head = &xlist[k] ;
   pt = head ;
   for (i=0; i<nsnp; i++) {  
    if (i<(nsnp-1)) {
     ptnext = &xlist[xind[i+1]] ;
    }
    else ptnext = NULL ;
    pt = &xlist[xind[i]] ;
    pt->next = ptnext ;
   }
 
   if (verbose) {
    for (pt = head; pt != NULL; pt = pt->next) {  
      printf("node: %5d  score: %9.3f\n",pos[pt->snum], pt->score) ;
    }
   }
/** 
 and build reverse pointers 
 tail record not needed.
*/

   for (pt = head; pt != NULL; pt = pt->next) {  
    ps = pt->next ;
    if (ps != NULL) ps->last = pt ;
   }

/**  
 now start main processing 
 cklist just checks list integrity 
 only does checks if debug==YES
*/
   cklist("M",head) ;
   nout = 0 ;
   pt = head ;
   for (;;) {  
    
    if (pt == NULL) break ;
    cklist("MX",pt) ;

    k = pt->snum ;
    if (xmark[k]==1) fatalx("list messed up\n") ;
    if (xmark[k]==-1) { 
     remlist(pt,head) ;
     pt=pt->next ;
     continue ;
    }
    cklist("A0",pt) ;
/** 
 get list of neighbors within adblim
*/
    getndlist(nbuff, &nsize, adblim, nsnp, k) ;
/**
     if (pos[k] == 39139589) { 
        for (i=0; i<nsize; i++) {  
         j = nbuff[i] ;
         printf("yy1a: %d %d %d\n", j, pos[j], xmark[j]) ;
        }
      }
*/
    cklist("A1",pt) ;
/** 
 and prune to SNP's marked with 1 (used) 
*/
    mknlist2(nbuff, &nsize, xmark) ;
/**
     if (pos[k] == 39139589) { 
        for (i=0; i<nsize; i++) {  
         j = nbuff[i] ;
         printf("yy1b: %d %d\n", j, pos[j]) ;
        }
      }
*/
    cklist("A2",pt) ;
    if (nsize>KMAX) {
     printf("##not using %5d... too many neighbors\n",k) ;
     xmark[k] = -2 ; 
     remlist(pt,head) ;
     pt = pt->next ;
     continue ;
    }
    if (nsize != pt->npt) {  
/**      
 new neighbor.  need to recompute mut. inf. 
*/
      nbuff[nsize] = k ;
      y  = mutinfd(nbuff,nsize+1,cfreq) ;
/**
      if (pos[k] == 39139589) { 
       printf("zz1 %4d %9.3f\n", k, y) ;
        for (i=0; i<nsize; i++) {  
         j = nbuff[i] ;
         printf("zz1b: %d %d\n", j, pos[j]) ;
        }
      }
*/
      if (pt->score > 1.5) y= 2.0 - 0.01 * (double) (nsize+1) ;
      if (y>pt->score+.001)  {
       /* funky kmax stuff */
       printf("warning  inf increases %12.6f %12.6f\n", y, pt->score) ;
       y = pt -> score ;
      }
      ps = pt->last ;
      remlist(pt,head) ;
      pt -> npt = nsize ;
      pt -> score = y ;
      inslist(pt,head) ;
      pt = ps->next ;
      continue ;
    }
    xmark[k] = 1 ;
    tdis = gendis[k] ;
    if (usephys) tdis = 0.0 ;
    fprintf(ofile,"%12d %9.3f %9.3f %2d  %9.3f %9.3f %9.3f %15s %s\n",
     pos[k], tdis, 
     pt->score, nsize,xinf[k], p1[k],p2[k],pplace[k], tags[k]) ;
    cklist("M0",pt) ;
/**     
 now get list of close neighbors and knock em out   
*/
    getnlist(nbuff, &nsize, noclose, nsnp, k) ;
    for (i=0; i<nsize; i++) {  
     kk = nbuff[i] ; 
     if (xmark[kk] == 0) { 
       xmark[kk] = -1 ;
       if (verbose) 
        printf("killing(pos) %d  %d pos: %d  %d gd: %9.3f %9.3f %d\n", 
        kk, k, pos[kk], pos[k], gendis[kk], gendis[k], noclose) ;
      }
    }
    getndlist(nbuff, &nsize, noclosegd, nsnp, k) ;
    for (i=0; i<nsize; i++) {  
     kk = nbuff[i] ; 
//   if (pos[kk] == 34510678) printf("zzbug %d %d  %d %d\n",k,kk,pos[k], pos[kk]) ;
     if (xmark[kk] == 0) {
       xmark[kk] = -1 ;
       if (verbose) 
        printf("killing(gen) %d  %d pos: %d  %d gd: %9.3f %9.3f\n", kk, k, pos[kk], pos[k], gendis[kk], gendis[k]) ;
     }
    }
    pt = pt->next ;
   }

   return 0 ;

}
void cklist(char *str,struct node *pfix ) 
{
  struct node *qt, *qtl ;
  double lscore ;
  int nloop = 0 ;

  if (pfix==NULL) return ;
  if (debug == NO) return ;

  lscore = pfix->score ;
  for (qt = pfix; qt!=NULL; qt = qt->next) {
   ++nloop ;
   if (nloop>5000) fatalx("cklist looping (%s)\n",str) ;
   if (isnan(qt->score)) fatalx("cklist (%s) fails\n",str) ;
   if (verbose) 
   if (lscore<qt->score) fatalx("cklists (%s) fails\n",str) ;
   lscore = qt->score ;
   qtl = qt->last ;
   if ((qtl != NULL) && (qtl->next != qt)) 
     fatalx("cklist mismatch (%s) fails\n",str) ;
  }
}
void remlist(struct node *pfix, struct node *head) 
{
  struct node *px,  *ps, *pt, *qt ;
  cklist("RX",pfix) ;
  pt = pfix ;
  px = pfix->next ;
  ps = pfix->last ;
  if (ps != NULL) ps->next = px ;
  if (px != NULL) px->last = ps ;
  cklist("R",head) ;
/** the following loop is just debug */
  for (qt=head; qt!=NULL; qt=qt->next) {  
   if (qt == pfix) fatalx("remlist dup\n") ;
  }
}
void inslist(struct node *pfix, struct node *head) 
/** 
 pfix is out of order in linked list. Fix
*/
{
  struct node *px, *pt,  *ps,  *qt, *qtl ;
  double lscore ;
  int nloop = 0 ;

  cklist("IX",head) ;

  pt = pfix ;
  px = head  ;

  if (px==NULL) return ;
  if (pt->score > px->score) fatalx("head score too small\n") ;


  lscore = px->score ;
  for (qt = px; qt!=NULL; qt = qt->next) {
   ++nloop ;
   if (nloop>100000) fatalx("inslist looping\n") ;
   if (pt->score >= qt->score) break ;
   if (qt->score>lscore) fatalx("bad scorelist\n") ;
   qtl = qt ;
   lscore = qt->score  ;
  }

  px = qt  ;
  ps = qtl ;

  cklist("I0",head) ;
  cklist("I1",px) ;
  if (ps !=NULL) ps->next = pt ;
  if (px !=NULL) px->last = pt ;
  pt -> next = px ;
  pt -> last = ps ;
  cklist("I",head) ;
}

void
mknlist2(int *nbuff, int *nsize, int *xmark)  
/* prune nlist to indices marked with 1 */
{
   int *tbuff, nt, nout=0 ;
   int i, k ;
   nt = *nsize ;

   if (nt==0) return ;
   ZALLOC(tbuff, nt, int) ;
   for (i=0; i<nt; i++) {
    k = nbuff[i] ;
    if (xmark[k] == 1)  {  
     tbuff[nout] = k ;
     ++nout ;
    }
   }
   copyiarr(tbuff, nbuff, nout) ;
   *nsize = nout ;
   free(tbuff) ;
}


void getndlist(int *nlist, int *nsize, double dis, int nsnp, int k)  
/** 
 return a list of snps within dis of k
 k is NOT returned 
*/

{
     int lo, hi, n, t ;
     t =  lo  = k ;

     if (dis<=0.0) {  
      *nsize = 0; 
      return ; 
     }
     
     for (;;) {
     if ( (t<0) || ((gendis[k]-gendis[t]) > dis))  break ;
      lo = t ;
      t-- ;
     }

     t = hi = k ;
     for (;;) {
     if ( (t>=nsnp) || ((gendis[t]-gendis[k]) > dis))  break ;
      hi = t ;
      t++ ;
     }
     n = 0;
     for (t=lo; t<=hi; t++) {  
      if (t != k) {  
       nlist[n] = t ;
       ++n ;
      }
     }
     *nsize = n ;
}
void getnlist(int *nlist, int *nsize, int dis, int nsnp, int k)  
/** 
 return a list of snps within dis of k
 k is NOT returned 
*/

{
     int lo, hi, n, t ;
     t =  lo  = k ;
     
     for (;;) {
     if ( (t<0) || ((pos[k]-pos[t]) > dis))  break ;
      lo = t ;
      t-- ;
     }

     t = hi = k ;
     for (;;) {
     if ( (t>=nsnp) || ((pos[t]-pos[k]) > dis))  break ;
      hi = t ;
      t++ ;
     }
     n = 0;
     for (t=lo; t<=hi; t++) {  
      if (t != k) {  
       nlist[n] = t ;
       ++n ;
      }
     }
     *nsize = n ;
}
void getdata(FILE *ifile, int *pos, double *gendis, 
  double *p0, double *p1, 
  char **pplace,  int *num, char **tags, int *forceit) 
{
   int nsplit ;
   char line[MAXSTR] ;
   char *spt[MAXFIELD] ;
   char *tag ;
   int  tnum=0 ;
   while (fgets(line, MAXSTR,ifile) != NULL) {
    nsplit = splitup(line, spt, MAXFIELD) ;
    if ((nsplit<3) || (spt[0][0] == '#')) { 
     freeup(spt, nsplit) ;  continue ;
    }
    if (tnum>=MAXNUM) fatalx("MAXNUM too small\n") ;
    p0[tnum] = atof(spt[3]) ;
    p1[tnum] = atof(spt[4]) ;
    pos[tnum] = atoi(spt[1]) ;
    gendis[tnum] = atof(spt[2]) ;
    pplace[tnum] = strdup(spt[5]) ;
    tag = spt[7] ;
    tags[tnum] = strdup(tag) ;
    if ((tagforce != NULL) && (strcmp(spt[5], tagforce) == 0)) 
     forceit[tnum] = YES ;
    ++tnum ;
    freeup(spt, nsplit) ;
   }
   *num = tnum ;
}
double mutxx(double *dd, int m, int n) 
{
   double *aa, *bb ; 
   double y ;
   static int ncall = 0 ;

   ZALLOC(aa, m, double) ;
   ZALLOC(bb, n, double) ;

   rrsum(dd, bb, m, n) ;
   ccsum(dd, aa, m, n) ;

   if (ncall < -1) {
    printf("m: %d n: %d\n",m,n) ;
    printmat(dd,m,n) ;
    printf("\n") ;
    printmat(aa,1,m) ;
    printmat(bb,1,n) ;
   }
   y = entrop(aa,m) + entrop(bb,n) - entrop(dd,m*n) ;
   y = MAX(y, 0.0) ;

  free(aa) ;  free(bb) ;
  ++ncall ;
  return y ;
}
void m3(double *xout, double a0, double a1, double pf) 
/** 
 pf is prob of C  
 a0 is prob (0 | C) 
 a1 is prob (0 | A) 
*/
{
   double x0[2], x1[2], y0[2] ;
   double yy[3] ;

   x0[0] = a0 ;
   x0[1] = 1.0-a0 ;

   x1[0] = a1 ;
   x1[1] = 1.0-a1 ;

   y0[0] = pf ;
   y0[1] = 1.0-pf ;

   m2(yy,y0,y0,1.0) ;  /* prob of 0, 1, 2 C */
   m2(xout, x0, x0, yy[0]) ;
   m2(xout+3, x0, x1, yy[1]) ;
   m2(xout+6, x1, x1, yy[2]) ;
}
void m2(double *x, double *a, double *b, double scal)
{

  x[0] = a[0]*b[0] ;
  x[1] = a[0]*b[1] + a[1]*b[0] ;
  x[2] = a[1]*b[1] ; 
  vst(x, x, scal, 3) ;

}

double mutinfd(int *a, int n, double pf) 
/** 
 difference between mut inf 
 for n snps and n-1
*/

{
  double ans, oldans ;
  double *pp1, *pp2, *dis, gd ;
  int kmark , *alist, *ind, *b ;
  int k, i, j, aa, bb ;
  static int ncall = 0 ;
  double y ;

  if (n==0) return 0.0 ;
  if ((lambda==0.0)) { 
   ans = oldmutinf(a, n, pf) -  oldmutinf(a, n-1, pf)  ;
   return ans ;
  }
  ++ncall ;
  ZALLOC(dis, n, double) ;
  if (n>kmax) { 
/** just use kmax nearest distances */  
   ZALLOC(ind, n, int) ;
   bb = a[n-1] ;
   for (i=0; i<n-1; i++) {  
    aa = a[i] ;
    dis[i] = fabs(gendis[aa] - gendis[bb])  ;
   }
   sortit(dis, ind, n-1) ;
   ZALLOC(b, kmax, int) ;
   for (i=0; i<kmax-1; i++) { 
    j = ind[i] ;
    b[i] = a[j] ;
   }
   b[kmax-1] = bb ;
   free(ind) ;
   free(dis) ;
   y = mutinfd(b, kmax, pf) ;
   free(b) ;
   return y ;
  }
  ZALLOC(alist, n, int) ;
  ZALLOC(pp1, n, double) ;
  ZALLOC(pp2, n, double) ;

  copyiarr(a, alist, n) ;
  isortit(alist, NULL, n) ;
  kmark = -1 ;
  gd = -99999.0 ;
  for (i=0; i<n ; i++) { 
   k = alist[i] ;
   if (k == a[n-1]) kmark = i ;
   pp2[i] = p1[k] ;   /* dangerous bend */
   pp1[i] = p2[k] ;  
   dis[i] = gendis[k]*0.01 ; /* dangerous bend distances in cM in ranksnp */
   if (dis[i]<gd) fatalx("bad distance %9.3f %9.3f\n", gd, dis[i]) ;
   gd = dis[i] ;
  }
  if (kmark<0) fatalx("logic bug\n") ;
  ans = mutxhmmd(pp1, pp2, dis, n, kmark, theta, lambda)  ;

  if (ncall <= -1) { 
    oldans = mutinf(a, n, pf) - mutinf(a, n-1, pf) ;
    printf("zz %12.6f  %12.6f\n", ans, oldans) ;
  }
  return ans ;
}

double mutinf(int *a, int n, double pf) 
{
 if (n==0) return 0.0 ;
 if ((lambda==0.0) || (n==1)) { 
  return oldmutinf(a, n, pf) ;
 }
 return oldmutinf(a, n, pf) ;
}

double oldmutinf(int *a, int n, double pf) 
/** 

 given array of indices pointing to SNPs and 
 pf freqency returns mutual information about 
 SNP values and ancestry

*/
{
  double **fff ;
  double *xa, *xb ;
  double *xpt ;
  int i,l,k ;
  int nn ;
  double yy[3], y0[2] ;
  double y, ans  ;
  static int ncall=0 ;

   if (n==0) return 0.0 ;

   y0[0] = pf ;
   y0[1] = 1.0-pf ;

   m2(yy,y0,y0,1.0) ;  /* prob of 0, 1, 2 C */

  ZALLOC(fff, n, double *) ;

  nn = nnint (pow (3.0, (double) n) ) ;
  ZALLOC(xa, nn, double) ;
  ZALLOC(xb, nn*3, double) ;
  for (i=0; i<n; i++) {  
   ZALLOC(fff[i],9,double) ;
   k = a[i] ;
   m3(fff[i], p1[k], p2[k],pf) ; 
  }

  for (k=0; k<3; k++) {  
   l=1 ; xa[0] = 1.0 ;
   for (i=0; i<n; i++) { 
    xpt = fff[i] + 3*k ; 
    vst(xa+2*l, xa, xpt[2], l) ;
    vst(xa+1*l, xa, xpt[1], l) ;
    vst(xa+0*l, xa, xpt[0], l) ;
    l = 3*l ;
   }
   y = asum(xa, l) ;
   vst(xb+k*l, xa, yy[k]/y, l) ;
  }

  ans = mutxx(xb, l, 3) ;

  if (ncall< -1 ) {
    printf("zz check %d %9.3f\n", l, asum(xb, l*3) ) ;
    printf("%9.3f\n",asum(fff[0],9)) ;
    printf("%9.3f\n",asum(yy,3)) ;
    printf("%9.3f\n",ans) ;
  }


  free(xa) ;
  free(xb) ;
  for (i=0; i<n; i++) {  
   free(fff[i]) ;
  }
  free(fff) ;
  ++ncall ;
  return ans ;
}


