#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <nicklib.h>
#include <getpars.h>

#include "admutils.h"
#include "mcio.h"  
#include "simpsim.h" 
#include "mcmcpars.h" 

#define WVERSION   "105" 
/** 
 A simple coalescent simulator.  
 Multiple populations 
 Many unlinked markers
*/


#define MAXFL  50   
#define MAXSTR  512
#define MAXPOPS 50
#define MAXEVENT 50

char *trashdir = "/var/tmp" ;
int verbose = NO ;
int qtmode = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps = 1000, numindivs ; 

int globalpopsize = 10000 ;

double plo = .001 ;
double phi = .999 ;

char  *genooutfilename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indoutfilename = NULL ;
char  *indivname = NULL ;
char *eventfilename = NULL ;

char *outputname = NULL ;
FILE *ofile ;

int popsizes[MAXPOPS] ; 
int poplabels[MAXPOPS] ; 
int numpops, nplabel  ;
int xnumnode ;

EVENT **eventlist  ;
NODE **nodelist ;
NODE **initnodes ;

char  unknowngender = 'U' ;

void readcommands(int argc, char **argv) ;
void printind (char *indoutname, Indiv **indm, int numind)  ;
int eventtype (char *sx)  ;
void mergit1(EVENT *oldev, EVENT *ev, int lp1, int lp2)  ;
void mergit2(EVENT *oldev, EVENT *ev, NODE **ndlist)  ;
void nodeinit (NODE *nodept, int lpop, int color)  ;
void doevent(EVENT *event)  ; 
double cxmean(EVENT *event, int *popsamp, int lpop)  ; 
void printevent(EVENT *event)  ;

int colorit(NODE ** inode, int num)   ;
int docolor(NODE *node, double f)  ;

int loadevents (char *eventfilename) ;

int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **eglist ;
  int  lsnplist, lindlist, numeg ;
  int i,j, k, t, n, x, isnp, col; 
  SNP *cupt, *cupt1, *cupt2, *cupt3 ;
  Indiv *indx ;
  double gpos1,gpos2,cpos1,cpos2,gd, cd, gd100 ;
  double rthresh, zt ;
  int numvind  ;

  int ch1, ch2 ;
  int fmnum , lmnum ;
  int num, n1, n2, ninit ;

  int nindiv = 0, e, f, lag=1  ;
  double xc[9], xd[4], xc2[9] ;
  double ychi, zscore, zthresh = 20.0 ;
  int nignore, numrisks = 1 ;
  char str[MAXSTR]  ;
  NODE *nodept, *npt ;
  int numevent, ev ;
  int kode, iter ;

  ivclear(poplabels, -1, MAXPOPS) ;
  readcommands(argc, argv) ;

 nplabel = 0 ;
 for (i=0; i<numpops; i++) { 
  if (poplabels[i] < 0) poplabels[i] = i + 1  ;
  nplabel = MAX(nplabel, poplabels[i]) ;
 }
 ++nplabel ;

 numindivs = intsum(popsizes, numpops) ;
 printf("numsnps: %d  numindivs: %d\n", numsnps, numindivs) ;
 ZALLOC(snpmarkers, numsnps, SNP *) ;
 for (i=0; i<numsnps; i++) {  
  ZALLOC(snpmarkers[i], 1, SNP) ;
  cupt = snpmarkers[i] ;
  clearsnp(cupt) ;
  cupt -> ngtypes = numindivs ;
  ZALLOC(cupt -> gtypes, numindivs, int) ;
  ivclear(cupt -> gtypes, -1, numindivs) ;
  sprintf(cupt -> ID, "SNP:%d", i+1) ;
  cupt -> genpos = 0.01* (double) i ;
  cupt -> physpos = 10000 * i ;  
  cupt -> chrom = 1 ;
 }
  
  ZALLOC(indivmarkers, numindivs, Indiv *) ;
  for (i=0; i < numindivs; i++) {  
    ZALLOC(indivmarkers[i], 1, Indiv) ;
    indx = indivmarkers[i] ;
    sprintf(indx->ID,"%s:%d","IND",i+1) ;
  }
  clearind(indivmarkers, numindivs) ;
  t = 0 ;
  for (k = 0 ; k<numpops; ++k) {  
   n = popsizes[k] ;
   sprintf(str, "pop:%d", poplabels[k]) ;
   for (i=0; i<n; ++i)   {
    indx = indivmarkers[t] ;
    ++t  ;  
    indx -> affstatus = poplabels[k] ;
    indx -> egroup = strdup(str) ;
   }
  }

  printind(indoutfilename, indivmarkers, numindivs) ;
  printsnps(snpoutfilename, snpmarkers, numsnps, 
   indivmarkers, NO, NO) ;

  t = numindivs*10 ;
  ZALLOC(nodelist, t, NODE *) ;
  for (x=0; x<t; ++x) { 
   ZALLOC(nodelist[x], 1, NODE) ;
   nodept = nodelist[x] ;
   nodeinit(nodept, -1, -1) ;
  }

  numevent = loadevents(eventfilename) ;

  ninit = t = 2*intsum(popsizes, numpops) ;
   
  ZALLOC(initnodes, t, NODE *) ;
  for (x=0; x<t; ++x) { 
   ZALLOC(initnodes[x], 1, NODE) ;
  }

  for (isnp=0; isnp<numsnps; ++isnp)  {  

  iter = 0 ;
  for (;;) { 
   ++iter ;

   x = 0 ;  
   for (k=0; k<numpops; k++) { 
    for (i=0; i<2*popsizes[k]; ++i) { 
     nodept = nodelist[x] ;
     npt = initnodes[x] ;
     nodeinit(nodept, poplabels[k], -1) ;
     nodeinit(npt, poplabels[k], -1) ;
     npt -> parent = (struct NODE *) nodept ;
     ++x ; 
    }
   }
   xnumnode = x ;
   for (ev=0; ev<numevent; ++ev) {  
    doevent(eventlist[ev]) ;   
   }
   kode = colorit(initnodes, ninit) ;
   printf("kode: %d\n", kode) ;
   if (kode>0) break ;
  }
  cupt = snpmarkers[isnp] ;
  for (k=0; k< numindivs; ++k) {  
   col = 0 ;
   nodept = initnodes[2*k] ; 
   if (nodept -> color == 1) ++col ;
   nodept = initnodes[2*k+1] ; 
   if (nodept -> color == 1) ++col ;
   putgtypes(cupt, k, col) ;
  }
 }
  printdata(genooutfilename, indoutfilename, snpmarkers, indivmarkers, numsnps, numindivs, NO);

  printf("##end of run\n") ;
  return 0 ;
}
int colorit(NODE ** inode, int num)  
{
   double f ;  
   int n1=0, n2=0, x, i ;

   f = DRAND() ;  

   for (i=0; i<num; i++) { 
    x = docolor(inode[i], f) ;
    if (x==1) ++n1 ;
    if (x==2) ++n2 ;
   }

   if ((n1==0) || (n2==0)) return 0 ;
   return 1 ;

}
int docolor(NODE *node, double f) 
// recursively color the tree 
{
   int k, t ; 
   double y ;
   NODE *xnode ;
    
   k = node -> color ;
   if (k>0) return k ;  
   if (node -> parent == NULL)  {  
    y = DRAND() ;  
    t = 1 ; 
    if (y<f) t=2 ;
    node -> color = t ; 
    return t ;
   }
   xnode = (NODE *) node -> parent ;
   t = docolor(xnode, f) ;  
   node -> color = t ; 
   return t ;
}

void doevent(EVENT *event) 

{  
  double *cmeans ;   
  int **nodeplist ;
  int  *popsamp  ;
  NODE *nodept ;
  EVENT *lastevent, *tevent ;
  int i, j, x, lpop, etype, lp1, lp2 ;
  int t, tt, a, b, xa, xb ; 
  double y, *ctime, nexttime, time ;
  int cmin, cmax ;

  ZALLOC(nodeplist, nplabel,  int *) ;
  for (j=0; j<nplabel; j++) {  
   ZALLOC(nodeplist[j], xnumnode, int) ;
  }
  ZALLOC(cmeans, nplabel, double) ;
  ZALLOC(ctime, nplabel, double) ;
  vclear(ctime, 1.0e20, nplabel) ;
  ZALLOC(popsamp, nplabel, int) ;
  lastevent  = (EVENT *) event -> lastevent ;
  if (lastevent != NULL) {  
   copyiarr(lastevent -> effsizes, event -> effsizes, nplabel) ;
  }
  for (x=0; x<xnumnode; ++x) { 
   nodept = nodelist[x] ;
   if (nodept -> parent != NULL) continue ;
   lpop = nodept -> lpop ;
   t = popsamp[lpop] ; 
   nodeplist[lpop][t] = x ;
   ++popsamp[lpop] ;
  }

  printevent(event) ;

  etype = event -> etype ;
  if (etype == 1)  {  
   lpop = event -> lp1 ;
   event -> effsizes[lpop] = event -> esize ;
  }
  if (etype == 2)  {  
   lp1 = event -> lp1 ;
   lp2 = event -> lp2 ;
   if (lp1 == lp2) fatalx("bad event (type 2)\n") ;
   t = popsamp[lp1] ;
   xb = popsamp[lp2] ;
   for (i=0; i<t; i++)  {  
    a = nodeplist[lp1][i] ;
    nodeplist[lp2][xb] = a ;
    ++xb ;
    nodept = nodelist[a] ;
    nodept -> lpop = lp2 ;
   }
   popsamp[lp1] = 0 ;
   popsamp[lp2] = xb ;
  }
 

  time = event -> time ;
  nexttime = event -> nexttime ;
  for (;;)   { 
   if (time>=nexttime) break ;
   for (j=0; j<nplabel; j++) {  
    ctime[j] = nexttime + 1.0 ;
    if (popsamp[j] <=1 ) continue ;
    cmeans[j] = cxmean(event, popsamp, j) ;
    ctime[j] = ranexp()*cmeans[j] ;
   }
   vlmaxmin(ctime, nplabel, NULL, &lpop)  ;  
   time += ctime[lpop] ;
   if (time>=nexttime) break ;
   t = popsamp[lpop] ;
   pick2(t, &a, &b) ;
   xa = nodeplist[lpop][a] ;
   xb = nodeplist[lpop][b] ;
   nodelist[xa] -> parent =  (struct NODE *) nodelist[xb] ;
   nodeplist[lpop][a] = nodeplist[lpop][t-1] ;  
   --popsamp[lpop] ;   
// printf("coalesc %d %d %d %d\n", lpop, a, b, popsamp[lpop]) ;
  }

  time = nexttime ; 
  printf("end event %9.3f\n", time) ; 
  printimat(popsamp, 1, nplabel) ;
  printf("\n\n") ;
  
  for (j=0; j<nplabel; j++) {  
   free(nodeplist[j]) ;
  }
  free(nodeplist) ;
  free(cmeans) ;
  free(ctime) ;

}
double cxmean(EVENT *event, int *popsamp, int lpop) 
{
// mean coalescence time 

   int t, tt, plab ;
   double  y ;

   t = popsamp[lpop] ;  
   if (t<=1) return 1.0e8 ;
   tt = event -> effsizes[lpop] ;
   if (tt == 0) return 1.0e+8 ;
   y =  (double) (2*tt) ;
   y /= 0.5 * (double) (t*(t-1)) ;
// printf("yyy %d %d  %d %d %12.6f\n", lpop, t,   tt, y) ;
   return y ;
}

void nodeinit (NODE *nodept, int lpop, int color) 
{

/**
typedef struct { 
  int lpop ;
  int color ; 
  struct NODE *parent ;
} NODE ;
*/

  nodept -> lpop = lpop ;  
  nodept -> color = color ;
  nodept -> parent = NULL ;
}

void readcommands(int argc, char **argv) 

{
  int i;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n, kode ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

/**
DIR2:  /fg/nfiles/admixdata/ms2
SSSS:  DIR2/outfiles 
genotypename: DIR2/autos_ccshad_fakes
eglistname:    DIR2/eurlist  
output:        eurout
*/
   kode = getintss(ph, "popsizes:", popsizes, &n) ;
   if (kode<=0) fatalx("popsizes: comulsory\n") ;

   numpops = n ;  
   kode = getintss(ph, "poplabels:", poplabels, &n) ;
   if ( (kode>0) && (n != numpops))  fatalx("poplabels popsizes mismatch\n") ;
   getint(ph, "numsnps:", &numsnps) ;


   getstring(ph, "genooutfilename:", &genooutfilename) ;
   getstring(ph, "indoutfilename:", &indoutfilename) ;
   getstring(ph, "eventfilename:", &eventfilename) ;
   getstring(ph, "output:", &outputname) ;
   writepars(ph) ;
   closepars(ph) ;

}
void printind (char *indoutname, Indiv **indm, int numind)  
{
   FILE *fff ;
   int i ; 
   Indiv *indx ;
     
   if (indoutname == NULL) return ;
   openit(indoutname, &fff, "w")  ;

   for (i=0; i< numind ; ++i)  {  
    indx = indm[i] ;
    fprintf (fff,"%20s ", indx -> ID) ;
    fprintf(fff,"%c ", indx -> gender) ;
    fprintf(fff,"%20s\n", indx -> egroup) ;
  }
  fclose (fff) ;

}

int loadevents (char *eventfilename) 
{

 FILE *efile ;
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k, i ;
 EVENT *event, *lastevent = NULL ;
 EVENT *ebase ;
 SNPDATA *sdpt ;

 int nevent = 0 ;
 int *cpops, npops  ;
 int lpop, lp1, lp2, nsize ;

  ZALLOC(eventlist, MAXEVENT, EVENT *) ;


  ZALLOC(cpops, nplabel, int)  ;
  ivclear(cpops, globalpopsize, nplabel) ;
  npops = numpops ;


   for (i=0; i< MAXEVENT; i++) { 
    event = eventlist[i] ;
    ZALLOC(event , 1, EVENT) ;
    ZALLOC(event -> effsizes, nplabel, int) ;
    ZALLOC(event -> lpoplist, numpops, int) ;
    copyiarr(cpops, event -> effsizes, nplabel) ;
    copyiarr(poplabels,  event -> lpoplist, numpops) ; // merge1 overwrites
    eventlist[i] = event ;
   }

   ZALLOC(ebase, 1, EVENT) ;
   event = ebase  ;  
   ZALLOC(event -> effsizes, nplabel, int) ;
   ZALLOC(event -> lpoplist, numpops, int) ;
   copyiarr(cpops, event -> effsizes, nplabel) ;
   copyiarr(poplabels,  event -> lpoplist, numpops) ; // merge1 overwrites

   event -> eventnum = nevent ;
   event -> nexttime = -1  ;
   event -> time = 0 ;    
   event -> lastevent = NULL ;


   event -> etype = 0 ; // a sort of noop 
   lastevent  = event ;
   eventlist[nevent] = event ;

   ++nevent ;


  fflush(stdout) ;
  openit(eventfilename, &efile, "r") ;  

  while (fgets(line, MAXSTR, efile) != NULL)  {
   fflush(stdout) ;
   nsplit = splitup(line, spt, MAXFF) ; 

   if (nsplit == 0) continue ;

   sx = spt[0] ;

   if (sx[0] == '#')  { 
     freeup(spt, nsplit) ;
     continue ;
   }

   k = eventtype(sx) ;  
   if (k<0) { 
    printf("event?? %s\n", line) ;
    freeup(spt, nsplit) ;
    continue ;
   }

   event = eventlist[nevent] ;  

   event -> eventnum = nevent ;
   event -> lastevent =  (struct EVENT *) lastevent ;
   ++nevent ;
   event -> time = atoi(spt[1]) ;
   event -> nexttime = -1  ;
   lastevent -> nexttime = event -> time ;  

   for (i=1; i<nsplit; i++) {  
    event -> evals[i-1] = atof(spt[i]) ;
   }


   switch (k)  {  
    case 1:   
     lpop = atoi(spt[2]) ;
     nsize = atoi(spt[3]) ;
     cpops[lpop] = nsize ;
     copyiarr(cpops, event -> effsizes, nplabel) ;
     event -> lp1 = lpop ;  
     event -> esize = nsize ;
     break ;

    case 2:  
      lp1 = atoi(spt[2]) ;
      lp2 = atoi(spt[3]) ;
      mergit1(lastevent, event, lp1, lp2) ;
      break ;
      
    case 10: 
     break ;

   }
   event -> etype = k ;
   lastevent  = event ;
  }
  free(cpops) ;
  fclose(efile) ;
  return nevent ;
}
int eventtype (char *sx) 
{
  if (strcmp(sx, "popsize") == 0) return 1 ;
  if (strcmp(sx, "split") == 0) return 2 ;
  if (strcmp(sx, "lasttime") == 0) return 10 ;
  return -1 ;
}

void mergit1(EVENT *lastev, EVENT *ev, int lp1, int lp2) 
{
/* called at setup time */

  int n, nn, i, k ;
  n = lastev -> npops ;
  ev -> npops = n -1 ;  
  nn = 0  ;
  for (i=0; i< n; i++)  {  
   k = lastev -> lpoplist[i]  ;
   if (k==lp1) continue ;
   ev -> lpoplist[nn] = k ;
   ++nn ;
  }
  ev -> npops = nn  ;  
  copyiarr(lastev -> effsizes, ev -> effsizes, numpops) ;
  ev -> lp1 = lp1 ; 
  ev -> lp2 = lp2 ;
}


/**
typedef struct {
  int time ; 
  int nexttime ; 
  int eventnum ;   
  int etype ;
  int npops ; 
  int *lpoplist ; 
  int *effsizes ;
  int *sampsizes ;
  int **indpt ; 
  int lp1 ; 
  int lp2 ;
}  EVENT ;

typedef struct { 
  int lpop ;
  int color ; 
  struct NODE *parent ;
} NODE ;

*/


void printevent(EVENT *event) 
{
  printf("event: %d  type %d  time: %d nexttime: %d\n", 
   event-> eventnum, event -> etype, event -> time, event -> nexttime) ;
  printimat(event ->  lpoplist, 1, numpops) ;
  printimat(event ->  effsizes, 1, nplabel) ;
  printf("lpop1 lpop2: %d %d\n", event -> lp1, event -> lp2) ;
}
