#include <nicklib.h>
#include <math.h>

extern int verbose ;

double xpest(double **gg, int *gobs, int *na, int *nb,  
 int neq, double *ppa, double *ppb) ;

void mk2from3ml(double *xd, double *xc, double p, double pp)  ;
void mk2from2(double *xd, double *xc ) ;

void loadpprob(double *pprob, double pa, double pb) ;
void gen3(double *ww, double a, double b) ;

double xpest2like(double **gg, int *gobs, int *na, int *nb,  
 int *iscasearr, 
 int neq, double ppa, double ppb, double risk)  ;

double
 xpbayes(double **gg, int *gobs, int *na, int *nb,  
 int *iscase,  int n, double pmesh, double rmesh) 
{
   
   double *pval, *rvp, *rvm ;                   
   double *rv ;
   double x, y, pmax, rmax, risk ;
   double pa, pb, r ;
   double yscore, ybase, yrbase, yrscore, ymax ;
   double *ysc, *yscb, *riskvals, *scores  ;
   int np , nr, k, a1, a2, nrisk ;
   int i, isg ;
   int skipsc ;

   np = nnint(1.0/pmesh) + 1 ;
   nr = nnint(1.0/rmesh) + 2 ;
   ZALLOC(pval, np, double) ;
   ZALLOC(rvp, nr, double) ;
   ZALLOC(rvm, nr, double) ;
   pmax = 1.0-pmesh ;
   x = pmesh ;
   for (k=0; k<np; k++) { 
    if ((x-pmax)>pmesh/2) break ;
    pval[k] = x ;
    x += pmesh ;
//  printf("pval: %9.3f\n",x) ;
   }
   np = k ;

   rmax = 2.0  ;
   x = 1.0 ;
   for (k=0; k<nr; k++) { 
    if ((x-rmax)>rmesh/2) break ;
    rvp[k] = x ;
    rvm[k] = 1.0/x ;
//  printf("rval: %9.3 %9.3f\n",x,1/x) ;
    x += rmesh ;
   }
   nr = k  ;
/**
   printf("zz %d %d\n", np, nr) ;
   fflush(stdout) ;
*/
   ZALLOC(yscb, np*np, double) ;
   ZALLOC(ysc, 2*nr, double) ;
   ZALLOC(riskvals, 2*nr, double) ;
   ZALLOC(scores, 2*nr, double) ;
   vclear(scores, -40.0, 2*nr) ;
   for (a1=0; a1<np; a1++) {  
    for (a2=0; a2<np; a2++) {  
      pa = pval[a1] ;
      pb = pval[a2] ;
      y = yscb[a1*np+a2] = 
      xpest2like(gg, gobs, na, nb,  
      iscase,  n, pa, pb, 1.0) ;
//    printf("zz2 %d %d %9.3f\n", a1, a2, y) ;
    }
   }
   fflush(stdout) ;
   vmaxmin(yscb, np*np, &ymax, NULL) ;
   ybase = ymax ;
   vsp(yscb, yscb, -ymax, np*np) ;
   vexp(yscb, yscb, np*np) ;
   vst(yscb, yscb, 1.0/(double) np*np, np*np) ;
   yrbase = log(asum(yscb, np*np)) + ymax ;
   vclear(ysc, -40.0, nr*2) ;
   nrisk = 0 ;
   for (isg=0; isg <=1; ++isg) { 
     if (isg==0) rv = rvp ;
     if (isg==1) rv = rvm ;
     skipsc = NO ;
     for (i=0; i<nr; i++) {
       risk = rv[i] ;
       vclear(yscb, -40.0, np*np) ;
       for (a1=0; a1<np; a1++) {  
         for (a2=0; a2<np; a2++) {  
          pa = pval[a1] ;
          pb = pval[a2] ;
          y = yscb[a1*np+a2] = 
           xpest2like(gg, gobs, na, nb,  
             iscase,  n, pa, pb, risk)  ;
          y -= ybase ;
          if (verbose && (y> -10.0)) 
           printf("details: %9.3f %9.3f %9.3f   %9.3f\n", 
           pa, pb, risk,  y) ;
          }
        }
       vmaxmin(yscb, np*np, &ymax, NULL) ;
       vsp(yscb, yscb, -ymax, np*np) ;
       vexp(yscb, yscb, np*np) ;
       vst(yscb, yscb, 1.0/(double) np*np, np*np) ;
       yrscore = log(asum(yscb, np*np)) + ymax -yrbase ;
       if (skipsc) {  
         y =  scores[nrisk] = -40.0 ;
       }
       else {
        y =  scores[nrisk] = MIN(yrscore, 40.0) ;
       }
       if (y < -10.0)  skipsc = YES ;
       if (y<-20.0) 
       riskvals[nrisk] = risk ;
       ++nrisk ;
      }
     }

     if (verbose) {
      for (i=0; i<nrisk; i++) {  
       printf("zz %3d %9.3f %9.3f\n",
        i, riskvals[i], scores[i]) ;
      }
     }

     vexp(scores, scores, nrisk) ;
     vst(scores, scores, 1.0/(double) nrisk, nrisk) ;
  
     yrscore = log10(asum(scores, nrisk)) ;
   
     
     free(pval); free(rvp) ; free(rvm) ;
     free(ysc); free(yscb) ;
     free(riskvals); free(scores) ;

      return yrscore ;
}


double xpest(double **gg, int *gobs, int *na, int *nb,  
 int neq, double *ppa, double *ppb) 
// null is pa = pb = 0.5 
{
  double xx[2], ww[3] ;
  double ya[2], yb[2] ;
  double pprob[3*3] ;
  double g3[3*3] ;
  double g2[2*2], dd[4] ;
  double pa, pb, qa, qb ;
  double y, yp, ylike, ybase, ynull, yold ;
  int iter, numiter = 100, g, k ;
  int e, b ;

  floatit(ya, na, 2) ;
  floatit(yb, nb, 2) ;
  pa = qa = 0.5 ;
  pb = qb = 0.5 ;
  ynull = ya[0]*log(qa) + ya[1]*log(pa) ;
  ynull += yb[0]*log(qb) + yb[1]*log(pb) ;
  ww[0] = ww[2] = 1.0 ;
  ww[1]  = 2.0 ;
  bal1(ww, 3) ;
  vlog(ww, ww, 3) ;
  for (k=0; k<neq; k++) {  
   g = gobs[k] ;
   ynull += ww[g] ;
  }

  copyarr (ya, xx, 2) ;
  vsp(xx, xx, .001, 2) ;
  pa = xx[1]/asum(xx,2)  ;

  copyarr (yb, xx, 2) ;
  vsp(xx, xx, .001, 2) ;
  pb = xx[1]/asum(xx,2)  ;
  ylike = 0.0 ;
  
  for (iter=1 ; iter <= numiter; ++iter) {  
   loadpprob(pprob, pa, pb) ;
   qa = 1.0-pa ;
   qb = 1.0-pb ;
   vzero(g3, 9) ;
   vzero(g2, 4) ;
   vvp(g2+0, g2+0, ya, 2) ;
   vvp(g2+2, g2+2, yb, 2) ;
   ylike = ya[0]*log(qa) + ya[1]*log(pa) ;
   ylike += yb[0]*log(qb) + yb[1]*log(pb) ;
   for (k=0; k<neq; k++) {  
    g = gobs[k] ;
    if (g<0) continue ;
    vvt(ww, gg[k], pprob+3*g, 3) ;
    y = asum(ww,3) ;
    if (y<=0.0) { 
     printf("bad ww %9.3f %9.3f %9.3f\n", ww[0], ww[1], ww[2]) ;
     fatalx("bad xpest %d %d\n",  iter, k) ;
    }
    yp = bal1(ww, 3) ;
    ylike += log(yp) ;  
    b = 2-g ;
    for (e=0; e<=2 ; ++e)  {
     g3[3*e+g] += ww[e] ;    
    }
   }
   mk2from3ml(dd, g3, pa, pb) ;
   if (iter==1) {
    ybase = ylike ;
    yold = ybase -1.0 ;
   }
   mk2from2(g2, g2)  ;

   vvp(g2, g2, dd,4) ;
   pa = g2[0]/asum(g2+0,2)  ;
   pb = g2[2]/asum(g2+2,2)  ;

/**
   printf("iter: %5d  llike:  %9.3f %9.3f %9.3f\n", 
    iter, ylike-ybase, pa, pb) ;
*/ 

   if (ylike< (yold+.0001) )  break ;

   yold = ylike ;
  }

  *ppa = pa ;
  *ppb = pb ;
  return ylike-ynull ;

}

void loadpprob(double *pprob, double pa, double pb)
{
  double ww[3] ;
  int e, g ;
  gen3(ww, pa, pa) ; e=0; 
  for (g=0; g<=2; ++g)  {
   pprob[3*g+e] = ww[g] ;
  }
  gen3(ww, pa, pb) ; e=1; 
  for (g=0; g<=2; ++g)  {
   pprob[3*g+e] = ww[g] ;
  }
  gen3(ww, pb, pb) ; e=2; 
  for (g=0; g<=2; ++g)  {
   pprob[3*g+e] = ww[g] ;
  }
}

void gen3(double *w, double pa, double pb) 
{
 double qa, qb ;
 qa = 1.0-pa ;
 qb = 1.0-pb ;
 w[2] = pa*pb ;
 w[1] = pa*qb+qa*pb ;
 w[0] = qa*qb ;
}

void
mk2from3ml(double *xd, double *xc, double p, double pp) 
/* like mk2from3 but no random sample */
{
   double q, qq  ;
   double p1, p2 ;
   double y  ;

   int  c ;
   int  e1, e2, b1, b2, n1, n2 ;
   int  k, t ;
   double a, aa, b, bb, x1, x2, xx ;

   q = 1-p ; qq = 1-pp ;


   vzero(xd, 4) ;
   
   for (k=0; k<9; k++)  {  
     e1 = e2 = b1 = b2 = n1 = n2 = 0 ;
    switch (k)  {  
     case 0:  
       e1 = 0 ; b1 = 1; n1 = 2;  
       y = q*q ;
       break ;
     case 1:  
       e1 = 0 ; b1 = 0; n1 = 1;  
       e2 = 0 ; b2 = 1; n2 = 1;  
       y = 2*q*p ;
       break ;
     case 2:  
       e1 = 0 ; b1 = 0; n1 = 2;  
       y = p*p ;
       break ;
     case 3:  
       e1 = 0 ; b1 = 1; n1 = 1;  
       e2 = 1 ; b2 = 1; n2 = 1;  
       y = q*qq ;  
       break ;
     case 4:  
       y = p*qq + pp*q ;  
       break ;
     case 5:  
       e1 = 0 ; b1 = 0; n1 = 1;  
       e2 = 1 ; b2 = 0; n2 = 1;  
       y = p*pp ;  
       break ;
     case 6:  
       e1 = 1 ; b1 = 1; n1 = 2;  
       y = qq*qq;
       break ;
     case 7:  
       e1 = 1 ; b1 = 0; n1 = 1;  
       e2 = 1 ; b2 = 1; n2 = 1;  
       y = 2*qq*pp ;
       break ;
     case 8:  
       e1 = 1 ; b1 = 0; n1 = 2;  
       y = pp*pp ;
       break ;
    }
    t = 2*e1 + b1 ;  xd[t] += (double) n1*xc[k] ;
    t = 2*e2 + b2 ;  xd[t] += (double) n2*xc[k] ;
   }

   p1 = p*qq ; p2 = pp*q ;  
   xx =  xc[4] ;  

   x1 = (xx* p1)/(p1+p2) ;
   x2 = xx - x1 ;

   e1 = 0; b1=0; 
   e2 = 1; b2=1;

   t = 2*e1 + b1 ;  xd[t] += x1 ;
   t = 2*e2 + b2 ;  xd[t] += x1 ;

   e1 = 0; b1=1; 
   e2 = 1; b2=0;
   t = 2*e1 + b1 ;  xd[t] += x2 ;
   t = 2*e2 + b2 ;  xd[t] += x2 ;

}

void
mk2from2(double *xd, double *xc) 
{
   double tt[4] ;
   tt[0] = xc[1] ;
   tt[1] = xc[0] ;
   tt[2] = xc[3] ;
   tt[3] = xc[2] ;
   copyarr(tt, xd, 4) ;
}

double xpest2like(double **gg, int *gobs, int *na, int *nb,  
 int *iscasearr, 
 int neq, double ppa, double ppb, double risk) 
// null is pa = pb = 0.5 
{
  double xx[2], ww[3] ;
  double ya[2], yb[2] ;
  double pprob[3*3], pprob1[3*3] ;
  double *ppr ;
  double ypa[2], ypb[2] ;
  double yqa[2], yqb[2] ;
  double pa, pb, qa, qb ;
  double yp, ylike, ybase, ynull, yold ;
  double tt, bb ;
  int g, k ;
  int e, b, iscase ;

  floatit(ya, na, 2) ;
  floatit(yb, nb, 2) ;
  pa = qa = 0.5 ;
  pb = qb = 0.5 ;
  ynull = ya[0]*log(qa) + ya[1]*log(pa) ;
  ynull += yb[0]*log(qb) + yb[1]*log(pb) ;
  ww[0] = ww[2] = 1.0 ;
  ww[1]  = 2.0 ;
  bal1(ww, 3) ;
  vlog(ww, ww, 3) ;
  for (k=0; k<neq; k++) {  
   g = gobs[k] ;
   ynull += ww[g] ;
  }

  pa = ppa ; pb = ppb ;

  ypa[0] = pa ; ypb[0] = pb ;
  
  tt = pa*risk ; bb = (1.0-pa) + tt ; ypa[1] = tt/bb ; 
  tt = pb*risk ; bb = (1.0-pb) + tt ; ypb[1] = tt/bb ;

  vsp(yqa, ypa, -1.0, 2) ;
  vsp(yqb, ypb, -1.0, 2) ;
  vst(yqa, yqa, -1.0, 2) ;
  vst(yqb, yqb, -1.0, 2) ;
  
   loadpprob(pprob, ypa[0], ypb[0]) ;
   loadpprob(pprob1, ypa[1], ypb[1]) ;

   qa = 1.0-pa ;
   qb = 1.0-pb ;
   ylike = ya[0]*log(qa) + ya[1]*log(pa) ;
   ylike += yb[0]*log(qb) + yb[1]*log(pb) ;
   for (k=0; k<neq; k++) {  
    g = gobs[k] ;
    iscase = iscasearr[k] ;
    ppr = pprob ;
    if (iscase==1) ppr = pprob1 ;
    vvt(ww, gg[k], ppr+3*g, 3) ;
    yp = bal1(ww, 3) ;
    ylike += log(yp) ;  
    if (k==-1) { 
     printf("zzq %d %9.3f %9.3f", iscase, risk, yp) ;  
     printmat(ww,1,3) ;
     printmat(gg[k],1,3) ;
     printmat(ppr,3,3) ;
    }
   }


  return ylike ;

}

